#include "GEMFReader.h"

void GEMFReader::Initialise(QString location)
{
	if ( ! ReadGEMFHeader(location))
	{
		qDebug() << "Could not read header.";
		Close();
	}
}

#define U32SIZE 4
#define U64SIZE 8

QHash<quint32,QString> GEMFReader::GetSourceList()
{
	return hshSources;
}

Coverage GEMFReader::RangeToCoverage(GEMFRange rs, quint32 zoom)
{
	Tile OSMTile;
	Coverage c;

	OSMTile.indexX = rs.xmin;
	OSMTile.indexY = rs.ymin;
	c.TopLeft = Mercator::OSMTileLatLonBounds(OSMTile, zoom, m_TileSize).first;

	OSMTile.indexX = rs.xmax;
	OSMTile.indexY = rs.ymax;
	c.BottomRight = Mercator::OSMTileLatLonBounds(OSMTile, zoom, m_TileSize).second;

	return c;
}

QList<Coverage> GEMFReader::GetMapCoverage(quint32 source_index, quint32 zoom)
{
	QList<Coverage> result;

	GEMFRange rs;
	foreach (rs, lstRangeData)
	{
		if ((rs.source_index == source_index)
				&& (rs.zoom == zoom))
		{
			result.append(RangeToCoverage(rs, zoom));
		}
	}
	return result;
}

void GEMFReader::SelectSource(quint32 source)
{
	if (hshSourceStates.contains(source))
	{
		quint32 this_source;
		foreach (this_source, hshSourceStates.keys())
		{
			if (this_source == source)
				EnableSource(this_source);
			else
				DisableSource(this_source);
		}
	}
}

void GEMFReader::EnableSource(quint32 source, bool bEnabled)
{
	if (hshSourceStates.contains(source))
	{
		hshSourceStates.insert(source, bEnabled);
	}
}

void GEMFReader::DisableSource(quint32 source)
{
	EnableSource(source, false);
}

void GEMFReader::AcceptAnySource()
{
	quint32 source;
	foreach (source, hshSourceStates.keys())
	{
		EnableSource(source);
	}
}

quint64 GEMFReader::ArrayToLong(QByteArray array)
{
	quint64 result = 0;
	quint32 i;

	for (i=0;i<U64SIZE;i++)
	{
		result <<= 8;
		result |= array[i] & 0xFF;
	}

	return result;
}

bool GEMFReader::ReadULong(QFile &fh, quint64 &result)
{
	QByteArray ulongArray = fh.read(U64SIZE);
	if (ulongArray.count() != U64SIZE)
	{
		qDebug() << "Couldn't read word";
		return false;
	}
	result = ArrayToLong(ulongArray);
	return true;
}


quint32 GEMFReader::ArrayToInt(QByteArray array)
{
	quint32 result = 0;
	quint32 i;

	for (i=0;i<U32SIZE;i++)
	{
		result <<= 8;
		result |= array[i] & 0xFF;
	}

	return result;
}

bool GEMFReader::ReadUInt(QFile &fh, quint32 &result)
{
	QByteArray uintArray = fh.read(U32SIZE);
	if (uintArray.count() != U32SIZE)
	{
		qDebug() << "Couldn't read word";
		return false;
	}
	result = ArrayToInt(uintArray);
	return true;
}

void GEMFReader::Close()
{
	if (gemf_file.isOpen())
	{
		gemf_file.close();
	}
}

bool GEMFReader::IsOpen()
{
	if (gemf_file.isOpen())
	{
		return true;
	}
	return false;
}

QImage GEMFReader::GetImage(quint32 x_index, quint32 y_index, quint32 zoom)
{
	GEMFRange rs;
	GEMFRange this_range = {0,0,0,0,0,0,0};
	bool bFoundRange = false;
	bool bOk = true;

	if ( ! IsOpen())
	{
		return QImage();
	}

	foreach (rs, lstRangeData)
	{
		if ((zoom == rs.zoom)
				&& (x_index >= rs.xmin)
				&& (x_index <= rs.xmax)
				&& (y_index >= rs.ymin)
				&& (y_index <= rs.ymax)
				&& (hshSourceStates[rs.source_index]))
		{
			this_range = rs;
			bFoundRange = true;
			break;
		}
	}

	if ( ! bFoundRange)
		return QImage();

	quint32 num_y = this_range.ymax + 1 - this_range.ymin;
	x_index = x_index - this_range.xmin;
	y_index = y_index - this_range.ymin;
	quint64 offset = (x_index * num_y) + y_index;
	offset *= (U32SIZE + U64SIZE);
	offset += this_range.offset;

	bOk = gemf_file.seek(offset);
	quint64 data_file_offset;
	quint32 data_file_length;
	QByteArray data;
	QImage img;

	if ( bOk )
		bOk = ReadULong(gemf_file, data_file_offset);
	if ( bOk )
		bOk = ReadUInt(gemf_file, data_file_length);

	QFile *pDataFile = &gemf_file;

	if (data_file_offset > lstFileSizes[0])
	{
		quint32 index = 0;
		quint32 filelist_count = lstFileSizes.size();
		while ((index < (filelist_count-1))
				&& (data_file_offset > lstFileSizes[index]))
		{
			data_file_offset -= lstFileSizes[index];
			index += 1;
		}
		pDataFile = new QFile(lstFileNames[index]);
		if ( ! pDataFile->open(QIODevice::ReadOnly))
		{
			qDebug() << "Could not open file " << lstFileNames[index];
			return QImage();
		}
	}

	if ( bOk )
		bOk = pDataFile->seek(data_file_offset);
	if ( bOk )
	{
		data = pDataFile->read(data_file_length);
		if (((quint32) data.count()) != data_file_length)
			bOk = false;
	}
	if ( bOk )
		bOk = img.loadFromData(data);

	if ( bOk )
		return img;
	else
		return QImage();
}

QList<quint32> GEMFReader::GetZoomLevels()
{
	QSet<quint32> setZoomLevels;

	GEMFRange rs;
	foreach (rs, lstRangeData)
	{
		setZoomLevels.insert(rs.zoom);
	}

	QList<quint32> result = setZoomLevels.toList();
	qSort(result);

	return result;
}

QList<quint32> GEMFReader::GetZoomLevels(quint32 source_index)
{
	QSet<quint32> setZoomLevels;
	GEMFRange rs;
	foreach (rs, lstRangeData)
	{
		if (rs.source_index == source_index)
			setZoomLevels.insert(rs.zoom);
	}

	QList<quint32> result = setZoomLevels.toList();
	qSort(result);

	return result;
}

bool GEMFReader::ReadGEMFHeader(QString location)
{
	if ( ! location.endsWith(".gemf"))
	{
		return false;
	}
	else if (loadedGEMFFile != location)
	{
		bool bOk;
		
		// Load the GEMF Header
		loadedGEMFFile = "";
		lstRangeData.clear();
		hshSources.clear();
		hshSourceStates.clear();
		lstFileSizes.clear();
		lstFileNames.clear();

		gemf_file.setFileName(location);

		if ( ! gemf_file.open(QIODevice::ReadOnly))
		{
			qDebug() << "Could not open file " << location;
			return false;
		}

		qDebug() << "Retrieving file sizes";
		QFileInfo fInfo(gemf_file);
		quint32 index = 1;
		while (fInfo.exists())
		{
			lstFileSizes << fInfo.size();
			lstFileNames << fInfo.absoluteFilePath();
			QString nextFile = QString(gemf_file.fileName() + "-%1").arg(index);
			qDebug() << "Checking " << nextFile;
			fInfo = QFileInfo(nextFile);
			index += 1;
		}

		qDebug() << "Reading GEMF Version";
		quint32 gemf_version;
		bOk = ReadUInt(gemf_file, gemf_version);
		if (( ! bOk) || (gemf_version != 4))
		{
			qDebug() << "Wrong GEMF Version " << gemf_version;
			gemf_file.close();
			return false;
		}

		qDebug() << "Reading tile size";
		bOk = ReadUInt(gemf_file, m_TileSize);
		if (( ! bOk) || (m_TileSize != 256))
		{
			qDebug() << "Wrong tile size " << m_TileSize;
			gemf_file.close();
			return false;
		}
		qDebug() << "Reading source list";
		quint32 source_count;
		qDebug() << "Reading source count";
		bOk = ReadUInt(gemf_file, source_count);
		if ( ! bOk)
		{
			qDebug() << "Not OK";
			gemf_file.close();
			return false;
		}

		qDebug() << "Source count: " << source_count;

		for (quint32 i=0;i<source_count;i++)
		{
			qDebug() << "Reading source " << i;
			quint32 source_index;
			quint32 source_name_length;
			bOk = ReadUInt(gemf_file, source_index);
			if ( ! bOk)
			{
				qDebug() << "Could not read source index";
				gemf_file.close();
				return false;
			}
			bOk = ReadUInt(gemf_file, source_name_length);
			if ( ! bOk)
			{
				qDebug() << "Could not read source name length";
				gemf_file.close();
				return false;
			}
			qDebug() << "Source Name Length: " << source_name_length;
			QString source_name = gemf_file.read(source_name_length);
			if (((quint32) source_name.count()) != source_name_length)
			{
				qDebug() << "Could not read source name";
				gemf_file.close();
				return false;
			}

			hshSources.insert(source_index, source_name);
			hshSourceStates.insert(source_index, true);
		}

		qDebug() << "Reading number of ranges";
		quint32 num_ranges;
		bOk = ReadUInt(gemf_file, num_ranges);
		if ( ! bOk)
		{
			qDebug() << "Not OK";
			gemf_file.close();
			return false;
		}
		qDebug() << "Number of ranges:" << num_ranges;

		for (quint32 i=0;i<num_ranges;i++)
		{
			qDebug() << "Reading range " << i;
			GEMFRange rs;
			bOk = ReadUInt(gemf_file, rs.zoom);
			if ( ! bOk)
			{
				qDebug() << "Could not read zoom level";
				gemf_file.close();
				return false;
			}
			bOk = ReadUInt(gemf_file, rs.xmin);
			if ( ! bOk)
			{
				qDebug() << "Could not read xmin";
				gemf_file.close();
				return false;
			}
			bOk = ReadUInt(gemf_file, rs.xmax);
			if ( ! bOk)
			{
				qDebug() << "Could not read xmax";
				gemf_file.close();
				return false;
			}
			bOk = ReadUInt(gemf_file, rs.ymin);
			if ( ! bOk)
			{
				qDebug() << "Could not read ymin";
				gemf_file.close();
				return false;
			}
			bOk = ReadUInt(gemf_file, rs.ymax);
			if ( ! bOk)
			{
				qDebug() << "Could not read ymax";
				gemf_file.close();
				return false;
			}
			bOk = ReadUInt(gemf_file, rs.source_index);
			if ( ! bOk)
			{
				qDebug() << "Could not read source index";
				gemf_file.close();
				return false;
			}
			bOk = ReadULong(gemf_file, rs.offset);
			if ( ! bOk)
			{
				qDebug() << "Could not read offset";
				gemf_file.close();
				return false;
			}

			qDebug() << "Read range: " 
				<< "Zoom = " << rs.zoom << "; "
				<< "XMin = " << rs.xmin << "; "
				<< "XMax = " << rs.xmax << "; "
				<< "YMin = " << rs.ymin << "; "
				<< "YMax = " << rs.ymax << "; "
				<< "Source = " << rs.source_index << "; "
				<< "Offset = " << rs.offset;

			lstRangeData << rs;
		}
		
		loadedGEMFFile = location;

		qDebug() << "Loaded GEMF header successfully";
		return true;
	}
	else
	{
		return true;
	}
}

