#include "MaverickReader.h"

void MaverickReader::Initialise(QString location)
{
	QFileInfo mapFI(location);

	if ( ! mapFI.exists())
	{
		return;
	}
	else if ( ! mapFI.isDir() )
	{
		return;
	}
	mapDir = QDir(location);

	quint32 index = 0;
	// Try each folder in the map store
	// as there should be subdirectories for layers
	QStringList lstLayers = mapDir.entryList();
	QString layer;
	hshSources.clear();
	hshSourceStates.clear();

	foreach (layer, lstLayers)
	{
		if (layer.startsWith("."))
			continue;
		QFileInfo fh(mapDir.absoluteFilePath(layer));
		if (fh.isDir())
		{
			hshSources.insert(index, layer);
			hshSourceStates.insert(index, true);
			index++;
		}
	}

	AcceptAnySource();

	m_bIsOpen = true;
	m_TileSize = 256;
}

void MaverickReader::Close()
{
	m_bIsOpen = false;
}

bool MaverickReader::IsOpen()
{
	return m_bIsOpen;
}

QHash<quint32,QString> MaverickReader::GetSourceList()
{
	return hshSources;
}

void MaverickReader::SelectSource(quint32 source)
{
	if (hshSourceStates.contains(source))
	{
		quint32 this_source;
		foreach (this_source, hshSourceStates.keys())
		{
			if (this_source == source)
				EnableSource(this_source);
			else
				DisableSource(this_source);
		}
	}
}

void MaverickReader::EnableSource(quint32 source, bool bEnabled)
{
	if (hshSourceStates.contains(source))
	{
		hshSourceStates.insert(source, bEnabled);
	}
}

void MaverickReader::DisableSource(quint32 source)
{
	EnableSource(source, false);
}

void MaverickReader::AcceptAnySource()
{
	quint32 source;
	foreach (source, hshSourceStates.keys())
	{
		EnableSource(source);
	}
}

QList<Coverage> MaverickReader::GetMapCoverage(quint32 source_index, quint32 zoom)
{
	QList<Coverage> result;
	
	if ( ! hshSources.contains(source_index))
		return result;

	QString layer = hshSources[source_index];
	QDir layerDir(mapDir.absoluteFilePath(layer));

	QStringList lstZoomEntries = layerDir.entryList();
	QString entry;
	bool bOk;

	foreach (entry, lstZoomEntries)
	{
		quint32 i = entry.toUInt(&bOk);
		if ( ! bOk)
			continue;
		if (i == zoom)
		{
			// Found the zoom level
			QDir zoomDir(layerDir.absoluteFilePath(entry));

			QStringList lstXTiles = zoomDir.entryList();
			QHash<quint32, QSet<quint32> > tileSets;

			QString x_str;
			foreach (x_str, lstXTiles)
			{
				quint32 x_val = x_str.toUInt(&bOk);
				if ( ! bOk)
					continue;
				QDir x_dir(zoomDir.absoluteFilePath(x_str));

				QStringList lstYTiles = x_dir.entryList();
				QSet<quint32> y_tiles;
				QString y_str;
				foreach (y_str, lstYTiles)
				{
					if (y_str.startsWith("."))
						continue;
					QString y_val_str = y_str.split(".", QString::SkipEmptyParts)[0];
					quint32 y_val = y_val_str.toUInt(&bOk);
					if ( ! bOk)
						continue;
					y_tiles.insert(y_val);
				}
				tileSets.insert(x_val, y_tiles);
			}

			// We now have the set of tiles
			QHash<QString, QSet<quint32> > UniqueSets;
			quint32 x;
			foreach (x, tileSets.keys())
			{
				QStringList strlst;
				quint32 y;
				foreach (y, tileSets[x])
				{
					strlst << QString::number(y);
				}
				QString strkey = strlst.join("-");

				if (UniqueSets.contains(strkey))
				{
					UniqueSets[strkey].insert(x);
				}
				else
				{
					QSet<quint32> s;
					s.insert(x);
					UniqueSets[strkey] = s;
				}
			}

			// We now have the unique sets
			QList<QHash<QString, quint32> > split_xsets;
			QSet<quint32> x_set;
			foreach (x_set, UniqueSets.values())
			{
				QList<quint32> x_list = x_set.toList();
				qSort(x_list);
				quint32 setxmin = x_list.first();
				quint32 setxmax = x_list.last();
				quint32 last_valid;
				bool isLastValid = false;

				for (quint32 xv=setxmin;xv<(setxmax+2);xv++)
				{
					if ( ( ! x_set.contains(xv) ) && (isLastValid))
					{
						QHash<QString, quint32> this_set;
						this_set.insert("xmin", last_valid);
						this_set.insert("xmax", xv-1);
						split_xsets.append(this_set);
						isLastValid = false;
					}
					else if ((x_set.contains(xv)) && ( ! isLastValid))
					{
						isLastValid = true;
						last_valid = xv;
					}
					else
					{
					}
				}
			}

			// We now have the X sets, so we need the y sets
			QList<QHash<QString, quint32> > full_sets;
			QHash<QString, quint32> xr;
			QSet<quint32> y_set;
			foreach (xr, split_xsets)
			{
				y_set = tileSets[xr["xmax"]];
				QList<quint32> y_list = y_set.toList();
				qSort(y_list);
				quint32 setymin = y_list.first();
				quint32 setymax = y_list.last();
				bool isLastValid = false;
				quint32 last_valid;

				for (quint32 yv=setymin;yv<(setymax+2);yv++)
				{
					if ( ( ! y_set.contains(yv) ) && (isLastValid))
					{
						QHash<QString, quint32> this_set;
						this_set.insert("xmin", xr["xmin"]);
						this_set.insert("xmax", xr["xmax"]);
						this_set.insert("ymin", last_valid);
						this_set.insert("ymax", yv-1);
						full_sets.append(this_set);
						isLastValid = false;
					}
					else if ((y_set.contains(yv)) && ( ! isLastValid))
					{
						last_valid = yv;
						isLastValid = true;
					}
					else
					{
					}
				}
			}

			// We now have full_sets, a list of sets of data in tile coordinates
			QHash<QString, quint32> single_range;
			foreach (single_range, full_sets)
			{
				Coverage c;
				Tile OSMTile;
				OSMTile.indexX = single_range["xmin"];
				OSMTile.indexY = single_range["ymin"];
				c.TopLeft = Mercator::OSMTileLatLonBounds(OSMTile, zoom, m_TileSize).first;
				OSMTile.indexX = single_range["xmax"];
				OSMTile.indexY = single_range["ymax"];
				c.BottomRight = Mercator::OSMTileLatLonBounds(OSMTile, zoom, m_TileSize).second;

				result << c;
			}
		}
	}

	return result;
}

QImage MaverickReader::GetImage(quint32 x_index, quint32 y_index, quint32 zoom)
{
	QImage img;

	// Try each folder in the map store
	// as there should be subdirectories for layers
	QString layer;

	QStringList filetypes;
	filetypes << "png" << "jpg" << "png.tile" << "jpg.tile";
	QString filetype;

	quint32 source_index;

	// First try the real zoom level
	foreach (source_index, hshSources.keys())
	{
		if ( ! hshSourceStates[source_index])
		{
			continue;
		}

		layer = hshSources[source_index];

		foreach (filetype, filetypes)
		{
			QString file_name = QDir::cleanPath(
					mapDir.absoluteFilePath(
						QString("%1/%2/%3/%4.%5")
						.arg(layer)
						.arg(zoom)
						.arg(x_index)
						.arg(y_index)
						.arg(filetype)
						));

			img = QImage(file_name);

			if ( ! img.isNull())
			{
				break;
			}
		}

		if ( ! img.isNull())
		{
			break;
		}
	}

	return img;

}


QList<quint32> MaverickReader::GetZoomLevels()
{
	QSet<quint32> setZoomLevels;
	QList<quint32> layerZoomLevels;

	quint32 source_index;

	// First try the real zoom level
	foreach (source_index, hshSources.keys())
	{
		if ( ! hshSourceStates[source_index])
		{
			continue;
		}

		layerZoomLevels = GetZoomLevels(source_index);
		setZoomLevels.unite(QSet<quint32>::fromList(layerZoomLevels));
	}

	QList<quint32> result = setZoomLevels.toList();
	qSort(result);
	return result;
}

QList<quint32> MaverickReader::GetZoomLevels(quint32 source_index)
{
	QSet<quint32> setZoomLevels;

	// This should be done for each layer (which exists in 
	// a subdirectory)
	QString layer;

	if ( ! hshSources.contains(source_index))
		return QList<quint32>();

	layer = hshSources[source_index];

	QDir layerDir(mapDir.absoluteFilePath(layer));

	QStringList lstEntries = layerDir.entryList();
	QString entry;
	bool bOk;
	foreach (entry, lstEntries)
	{
		quint32 i = entry.toUInt(&bOk);
		if (bOk)
		{
			setZoomLevels.insert(i);
		}
	}

	QList<quint32> result = setZoomLevels.toList();
	qSort(result);

	return result;
}

