/*
 * This code is very heavily based on the Python code from 
 *
 * http://www.maptiler.org/google-maps-coordinates-tile-bounds-projection/
 *
 * Since this is largely a (different language) copy of that code, here is the
 * original copyright notice from that file:
 *
 * Copyright (c) 2008 Klokan Petr Pridal. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
#include "Mercator.h"

LatLon Mercator::MetresToLatLon(Metres metres)
{
	LatLon result;

	result.longitude = (metres.mx / m_OriginShift) * 180.0;
	result.latitude = (metres.my / m_OriginShift) * 180.0;
	result.latitude  = (180.0 / M_PI) * (2.0 * atan(exp(result.latitude * M_PI / 180.0)) - (M_PI / 2.0));

	return result;
}

Metres Mercator::LatLonToMetres(LatLon latlon)
{
	Metres result;

	result.mx = (latlon.longitude * m_OriginShift) / 180.0;
	result.my = log(tan((90.0 + latlon.latitude) * M_PI / 360.0)) / (M_PI / 180.0);
	result.my *= m_OriginShift / 180.0;

	return result;
}

double Mercator::Resolution(quint32 zoom, quint32 tilesize)
{
	double init_resolution = 2 * M_PI * 6378137 / tilesize;
	return init_resolution / pow(2.0, zoom);
}

Pixels Mercator::MetresToPixels(Metres metres, quint32 zoom, quint32 tilesize)
{
	Pixels result;

	double res = Resolution(zoom, tilesize);

	result.px = (metres.mx + m_OriginShift) / res;
	result.py = (metres.my + m_OriginShift) / res;

	return result;
}

Metres Mercator::PixelsToMetres(Pixels pixels, quint32 zoom, quint32 tilesize)
{
	Metres result;
	double res = Resolution(zoom, tilesize);


	result.mx = (pixels.px * res) - m_OriginShift;
	result.my = (pixels.py * res) - m_OriginShift;

	return result;
}

Tile Mercator::PixelsToTMSTile(Pixels pixels, quint32 tilesize)
{
	Tile t;

	double tmsX = pixels.px / (1.0*tilesize);
	double tmsY = pixels.py / (1.0*tilesize);

	t.indexX = (quint32) (ceil(tmsX) - 1);
	t.indexY = (quint32) (ceil(tmsY) - 1);

	double prX = 1.0 * t.indexX * tilesize;
	double prY = 1.0 * t.indexY * tilesize;

	t.posX = (quint32) floor(pixels.px - prX);
	// Not sure about this:
	t.posY = (quint32) floor(tilesize + prY - pixels.py);

	return t;
}

QPair<Metres, Metres> Mercator::TMSTileMetreBounds(Tile t, quint32 zoom, quint32 tilesize)
{
	Pixels pc;
	pc.px  = t.indexX * tilesize;
	pc.py = t.indexY * tilesize;
	Metres min = PixelsToMetres(pc, zoom, tilesize);
	pc.px = (t.indexX+1) * tilesize;
	pc.py = (t.indexY+1) * tilesize;
	Metres max = PixelsToMetres(pc, zoom, tilesize);

	QPair<Metres, Metres> result;
	result.first = min;
	result.second = max;
	return result;
}

QPair<LatLon, LatLon> Mercator::TMSTileLatLonBounds(Tile t, quint32 zoom, quint32 tilesize)
{
	QPair<Metres, Metres> metre_bounds = TMSTileMetreBounds(t, zoom, tilesize);
	LatLon min = MetresToLatLon(metre_bounds.first);
	LatLon max = MetresToLatLon(metre_bounds.second);
	QPair<LatLon, LatLon> result;
	result.first = min;
	result.second = max;
	return result;
}

QPair<LatLon, LatLon> Mercator::OSMTileLatLonBounds(Tile t, quint32 zoom, quint32 tilesize)
{
	Tile tms = OSMTileToTMSTile(t, zoom);
	return TMSTileLatLonBounds(tms, zoom, tilesize);
}

Tile Mercator::TMSTileToOSMTile(Tile t, quint32 zoom)
{
	Tile result;

	result.indexX = t.indexX;
	result.indexY = (pow(2,zoom)-1) - t.indexY;

	result.posX = t.posX;
	result.posY = t.posY;

	return result;
}

Tile Mercator::OSMTileToTMSTile(Tile t, quint32 zoom)
{
	// Reversible translation:
	return TMSTileToOSMTile(t, zoom);
}

Tile Mercator::PixelsToOSMTile(Pixels pixels, quint32 zoom, quint32 tilesize)
{
	return TMSTileToOSMTile(PixelsToTMSTile(pixels, tilesize), zoom);
}

Tile Mercator::LatLonToOSMTile(LatLon latlon, quint32 zoom, quint32 tilesize)
{
	return PixelsToOSMTile(LatLonToPixels(latlon, zoom, tilesize), zoom, tilesize);
}

Pixels Mercator::LatLonToPixels(LatLon latlon, quint32 zoom, quint32 tilesize)
{
	return MetresToPixels(LatLonToMetres(latlon), zoom, tilesize);
}
LatLon Mercator::PixelsToLatLon(Pixels pixels, quint32 zoom, quint32 tilesize)
{
	return MetresToLatLon(PixelsToMetres(pixels, zoom, tilesize));
}

LatLon Mercator::WindowPosToLatLon(QPointF p, LatLon pageCentre, QSize winSize, quint32 zoom, quint32 tilesize)
{
	return PixelsToLatLon(WindowPosToPixels(p, pageCentre, winSize, zoom, tilesize), zoom, tilesize);
}

QPointF Mercator::LatLonToWindowPos(LatLon l, LatLon pageCentre, QSize winSize, quint32 zoom, quint32 tilesize)
{
	return PixelsToWindowPos(LatLonToPixels(l, zoom, tilesize), pageCentre, winSize, zoom, tilesize);
}

Pixels Mercator::WindowPosToPixels(QPointF p, LatLon pageCentre, QSize winSize, quint32 zoom, quint32 tilesize)
{
	QPointF centre(winSize.width() / 2.0, winSize.height() / 2.0);

	Pixels pixels = LatLonToPixels(pageCentre, zoom, tilesize);

	// Convert to point to the top left corner of the window
	pixels.px -= centre.x();
	pixels.py -= centre.y();
	pixels.px += p.x();
	pixels.py += p.y();

	return pixels;
}

QPointF Mercator::PixelsToWindowPos(Pixels p, LatLon pageCentre, QSize winSize, quint32 zoom, quint32 tilesize)
{
	Pixels p_centre = LatLonToPixels(pageCentre, zoom, tilesize);
	QPointF centre(winSize.width() / 2.0, winSize.height() / 2.0);

	Pixels p_offset;
	p_offset.px = p_centre.px - p.px;
	p_offset.py = p_centre.py - p.py;

	QPointF pt(centre.x() - p_offset.px, centre.y() + p_offset.py);

	return pt;
}

