#include <QtGui>
#include <QtCore>
#include <QDebug>

#include "QMapViewMainWindow.h"
#include "QMapViewScene.h"

QMapViewMainWindow::QMapViewMainWindow()
	: QMainWindow()
{
	qDebug() << "STARTING";
	ui.setupUi(this);

	m_bNoRecurse = false;
	connect(ui.actionExit, SIGNAL(triggered()),
			this, SLOT(close()));
	connect(ui.actionSelectTool, SIGNAL(toggled(bool)),
			this, SLOT(SelectToggled(bool)));
	connect(ui.actionPanTool, SIGNAL(toggled(bool)),
			this, SLOT(PanToggled(bool)));
	connect(ui.actionOpenMaverick, SIGNAL(triggered()),
			this, SLOT(OpenMaverick()));
	connect(ui.actionOpenGEMF, SIGNAL(triggered()),
			this, SLOT(OpenGEMF()));
	connect(ui.actionScaling, SIGNAL(toggled(bool)),
			ui.mapView, SLOT(SetScalingAllowed(bool)));
			
}

void QMapViewMainWindow::keyPressEvent(QKeyEvent *event)
{
	switch (event->key())
	{
		case Qt::Key_Plus:
			ui.mapView->ZoomIn();
			break;
		case Qt::Key_Minus:
			ui.mapView->ZoomOut();
			break;
		default:
			event->ignore();
			return;
	}
	event->accept();
}

void QMapViewMainWindow::SelectToggled(bool checked)
{
	if (m_bNoRecurse)
	{
		return;
	}
	m_bNoRecurse = true;

	ui.actionPanTool->setChecked(false);
	if ( ! checked)
	{
		ui.actionSelectTool->setChecked(true);
	}
	m_bNoRecurse = false;

	ui.mapView->SetMode(ModeSelect);
}

void QMapViewMainWindow::PanToggled(bool checked)
{
	if (m_bNoRecurse)
	{
		return;
	}

	m_bNoRecurse = true;

	ui.actionSelectTool->setChecked(false);
	if ( ! checked)
	{
		ui.actionPanTool->setChecked(true);
	}
	m_bNoRecurse = false;

	ui.mapView->SetMode(ModePan);
}

void QMapViewMainWindow::OpenMaverick()
{
	QString directory = QFileDialog::getExistingDirectory(this,
			tr("Select Maverick Tile Folder"));

	if (directory.isNull())
	{
		return;
	}

	ui.mapView->SetMapLocation(directory);

	CreateSourceListMenu();
	CreateCoverageMenu();
}

void QMapViewMainWindow::CoverageSelectionHandler()
{
	QAction *pAction = qobject_cast< QAction* >(QObject::sender());
	
	if ( ! pAction)
		return;

	QVariantHash actionData = pAction->data().toHash();

	ui.mapView->ShowCoverage(actionData["Source"].toInt(),
			actionData["Zoom Level"].toInt());
}

void QMapViewMainWindow::SourceSelectionHandler(bool state)
{
	QAction *pAction = qobject_cast< QAction* >(QObject::sender());

	if (m_bNoRecurse)
		return;

	m_bNoRecurse = true;

	if ( ! pAction)
	{
		return;
	}

	quint32 source_index = pAction->data().toInt();

	ui.mapView->EnableSource(source_index, state);

	bool act_state = true;
	QList<QAction *> lstpActions = ui.mnuSources->actions();
	QAction *pThisAction;
	foreach (pThisAction, lstpActions)
	{
		if ( ! pThisAction->isChecked())
		{
			act_state = false;
			break;
		}
	}
	pAllSourceAction->setChecked(act_state);

	m_bNoRecurse = false;
}

void QMapViewMainWindow::AllSourceSelectionHandler(bool state)
{
	QAction *pAction = qobject_cast< QAction* >(QObject::sender());
	(void) state;

	if (m_bNoRecurse)
		return;

	m_bNoRecurse = true;

	if ( ! pAction)
	{
		return;
	}

	QList<QAction *> lstpActions = ui.mnuSources->actions();
	QAction *pThisAction;
	foreach (pThisAction, lstpActions)
	{
		pThisAction->setChecked(true);
	}

	ui.mapView->AcceptAnySource();

	m_bNoRecurse = false;
}

void QMapViewMainWindow::CreateCoverageMenu()
{
	ui.mapView->ClearCoverage();
	QHash<quint32, QList<quint32> > SourceZoomLevels =
		ui.mapView->GetZoomLevelsAndSources();

	ui.mnuCoverage->clear();
	if (SourceZoomLevels.count() > 0)
	{
		QString coverageName;
		QAction *pAction;
		pAction = new QAction("Clear Coverage", this);
		pAction->setData("CLEAR");
		connect(pAction, SIGNAL(triggered()),
				ui.mapView, SLOT(ClearCoverage()));
		ui.mnuCoverage->addAction(pAction);
		ui.mnuCoverage->addSeparator();
		quint32 zoom_level;
		quint32 source_index;
		QHash<quint32, QString> hshSources = ui.mapView->GetSourceList();

		foreach (source_index, SourceZoomLevels.keys())
		{
			if ( ! hshSources.contains(source_index))
			{
				qDebug() << "Key missing error";
				continue;
			}

			QList<quint32> zoomLevels = SourceZoomLevels[source_index];

			QMenu *pSourceMenu = new QMenu(hshSources[source_index], this);

			foreach (zoom_level, zoomLevels)
			{
				QVariantHash actionData;
				actionData.insert("Source", source_index);
				actionData.insert("Zoom Level", zoom_level);

				pAction = new QAction(QString("Zoom Level %1").arg(zoom_level), this);
				pAction->setData(actionData);
				connect(pAction, SIGNAL(triggered()),
						this, SLOT(CoverageSelectionHandler()));
				pSourceMenu->addAction(pAction);
			}
			ui.mnuCoverage->addMenu(pSourceMenu);
		}
		ui.mnuCoverage->setEnabled(true);
	}
	else
	{
		ui.mnuCoverage->setEnabled(false);
	}
}

void QMapViewMainWindow::CreateSourceListMenu()
{
	ui.mapView->AcceptAnySource();
	QHash<quint32, QString> sources = ui.mapView->GetSourceList();

	ui.mnuSources->clear();
	if (sources.count() > 0)
	{
		QString sourceName;
		pAllSourceAction = new QAction("All sources", this);
		pAllSourceAction->setCheckable(true);
		pAllSourceAction->setChecked(true);
		pAllSourceAction->setData("All");
		connect(pAllSourceAction, SIGNAL(toggled(bool)),
				this, SLOT(AllSourceSelectionHandler(bool)));
		ui.mnuSources->addAction(pAllSourceAction);
		ui.mnuSources->addSeparator();
		QAction *pAction;
		quint32 source_index;
		foreach (source_index, sources.keys())
		{
			pAction = new QAction(sources[source_index], this);
			pAction->setCheckable(true);
			pAction->setChecked(true);
			pAction->setData(source_index);
			connect(pAction, SIGNAL(toggled(bool)),
					this, SLOT(SourceSelectionHandler(bool)));
			ui.mnuSources->addAction(pAction);
		}
		ui.mnuSources->setEnabled(true);
	}
	else
	{
		ui.mnuSources->setEnabled(false);
	}
}

void QMapViewMainWindow::OpenGEMF()
{
	QString gemf_file = QFileDialog::getOpenFileName(this,
			tr("Select GEMF Header File"),
			QString(),
			"GEMF Files (*.gemf);;All Files (*.*)");

	if (gemf_file.isNull())
	{
		return;
	}

	ui.mapView->SetMapLocation(gemf_file);

	CreateSourceListMenu();
	CreateCoverageMenu();
}
