#include <QtGui>
#include <cmath>

#include "QMapViewScene.h"
#include "MaverickReader.h"
#include "GEMFReader.h"

QMapViewScene::QMapViewScene(QWidget *parent)
	: QFrame(parent)
{
	SetTileSize(256);
	m_BlankImage = QImage(QSize(m_TileSize, m_TileSize), QImage::Format_RGB16);
	m_BlankImage.fill(Qt::black);
	m_ViewInfo.ZoomLevel = 0xFFFF;
	m_Mode = ModePan;

	double lat = 51.6805;
	double lon = -2.3547;
	LatLon l = {lat, lon};

	m_lstMarkerLocations << l;
	
	m_ViewInfo.PageCentre = l;

	m_bCanScale = false;

	m_Moving = false;

	pImgReader = NULL;

	SetMapLocation("");
}

void QMapViewScene::SetTileSize(quint32 tilesize)
{
	m_TileSize = tilesize;
}

void QMapViewScene::SetScalingAllowed(bool allowed)
{
	m_bCanScale = allowed;
}

QHash<quint32,QString> QMapViewScene::GetSourceList()
{
	if ((pImgReader == NULL) || (! pImgReader->IsOpen()))
		return QHash<quint32,QString>();
	else
		return pImgReader->GetSourceList();
}

void QMapViewScene::SelectSource(quint32 source)
{
	if ((pImgReader == NULL) || (! pImgReader->IsOpen()))
		return;
	else
		pImgReader->SelectSource(source);
	update();
}

void QMapViewScene::EnableSource(quint32 source, bool bEnabled)
{
	if ((pImgReader == NULL) || (! pImgReader->IsOpen()))
		return;
	else
		pImgReader->EnableSource(source, bEnabled);
	update();
}

void QMapViewScene::DisableSource(quint32 source)
{
	if ((pImgReader == NULL) || (! pImgReader->IsOpen()))
		return;
	else
		pImgReader->DisableSource(source);
	update();
}

void QMapViewScene::AcceptAnySource()
{
	if ((pImgReader == NULL) || (! pImgReader->IsOpen()))
		return;
	else
		pImgReader->AcceptAnySource();
	update();
}

void QMapViewScene::mousePressEvent(QMouseEvent *mouseEvent)
{
	if (mouseEvent->button() != Qt::LeftButton)
		return;

	QSize s = size();

	QPointF pos = mouseEvent->posF();

	if (m_Mode == ModePan)
	{
		Pixels pixels = Mercator::LatLonToPixels(m_ViewInfo.PageCentre,
				m_ViewInfo.ZoomLevel, m_TileSize);
		pixels.px -= (s.width() / 2.0);
		pixels.py -= (s.height() / 2.0);
		pixels.px += pos.x();
		pixels.py += pos.y();

		m_StartPoint = pos;
		m_StartCentre = Mercator::LatLonToPixels(m_ViewInfo.PageCentre,
				m_ViewInfo.ZoomLevel, m_TileSize);

		LatLon l = Mercator::PixelsToLatLon(pixels,
				m_ViewInfo.ZoomLevel, m_TileSize);

		qDebug() << "Clicked " << l.latitude << "," << l.longitude;

		m_Moving = true;
	}
	else if (m_Mode == ModeSelect)
	{
		m_SelectionRect.setTopLeft(pos);
		m_SelectionRect.setBottomRight(pos);

		m_Selecting = true;
	}

	QWidget::mousePressEvent(mouseEvent);
}

void QMapViewScene::mouseMoveEvent(QMouseEvent *mouseEvent)
{
	if (m_Moving)
	{
		QPointF pos = mouseEvent->posF();

		Pixels pc = m_StartCentre;
		pc.px += m_StartPoint.x() - pos.x();
		pc.py += pos.y() - m_StartPoint.y();
		m_ViewInfo.PageCentre = Mercator::PixelsToLatLon(pc,
				m_ViewInfo.ZoomLevel, m_TileSize);

		update();
	}
	else if (m_Selecting)
	{
		QPointF pos = mouseEvent->posF();
		m_SelectionRect.setBottomRight(pos);

		update();
	}
	QWidget::mouseMoveEvent(mouseEvent);
}

Tile QMapViewScene::FindTileContaining(LatLon l, quint32 zoom)
{
	return Mercator::LatLonToOSMTile(l, zoom, m_TileSize);
}

QHash<quint32, QList<quint32> > QMapViewScene::GetZoomLevelsAndSources(void)
{
	QHash<quint32, QList<quint32> > details;

	if ((pImgReader == NULL) || (! pImgReader->IsOpen()))
	{
		return details;
	}
	else
	{
		QHash<quint32, QString> hshSources = pImgReader->GetSourceList();
		quint32 source_index;
		QString source_name;

		foreach (source_index, hshSources.keys())
		{
			QList<quint32> lstZoomLevels = pImgReader->GetZoomLevels(source_index);
			details.insert(source_index, lstZoomLevels);
		}

		return details;
	}
}

void QMapViewScene::ShowCoverage(quint32 source_index, quint32 zoom)
{
	// Test code for now
	if ((pImgReader == NULL) || (! pImgReader->IsOpen()))
		return;

	QList<Coverage> mapCoverage = pImgReader->GetMapCoverage(source_index, zoom);

	lstCoverage.clear();
	lstCoverage.append(mapCoverage);
	update();
}

void QMapViewScene::ClearCoverage(void)
{
	lstCoverage.clear();
	update();
}

void QMapViewScene::mouseReleaseEvent(QMouseEvent *mouseEvent)
{
	if (m_Selecting && ( ! m_SelectionRect.isNull()))
	{
		QPointF tl = m_SelectionRect.topLeft();
		QPointF br = m_SelectionRect.bottomRight();

		LatLon l_tl = Mercator::WindowPosToLatLon(tl,
				m_ViewInfo.PageCentre, size(), m_ViewInfo.ZoomLevel, m_TileSize);
		LatLon l_br = Mercator::WindowPosToLatLon(br,
				m_ViewInfo.PageCentre, size(), m_ViewInfo.ZoomLevel, m_TileSize);

		QHash<quint32, QPair<Tile, Tile> > hshCorners;

		quint32 z;

		qDebug() << "tile_range = {}";
		for (z=m_MinZoom;z<=m_MaxZoom;z++)
		{
			Tile t_tl = FindTileContaining(l_tl, z);
			Tile t_br = FindTileContaining(l_br, z);

			Tile verified_tl = {0,0,0,0};
			Tile verified_br = {0,0,0,0};

			verified_tl.indexX = qMin(t_tl.indexX, t_br.indexX);
			verified_tl.indexY = qMin(t_tl.indexY, t_br.indexY);
			verified_br.indexX = qMax(t_tl.indexX, t_br.indexX);
			verified_br.indexY = qMax(t_tl.indexY, t_br.indexY);

			QPair<Tile, Tile> pair;
			pair.first = verified_tl;
			pair.second = verified_br;

			hshCorners.insert(z, pair);

			qDebug() << QString("tile_range[%1] = {'xmin': %2, 'xmax': %3, 'ymin': %4, 'ymax': %5}")
				.arg(z)
				.arg(pair.first.indexX)
				.arg(pair.second.indexX)
				.arg(pair.first.indexY)
				.arg(pair.second.indexY);
		}

	}

	m_Moving = false;
	m_Selecting = false;

	QWidget::mouseReleaseEvent(mouseEvent);
}

void QMapViewScene::SetMode(Mode m)
{
	m_Mode = m;
	m_SelectionRect = QRect();
}

void QMapViewScene::SetZoomLevel(quint32 zoom)
{
	if ((zoom >= m_MinZoom) && (zoom < m_MaxZoom))
	{
		m_ViewInfo.ZoomLevel = zoom;
	}
}

void QMapViewScene::ZoomIn()
{
	quint32 existingZoom = m_ViewInfo.ZoomLevel;

	for (quint32 newZoom=existingZoom+1;newZoom<=m_MaxZoom;newZoom++)
	{
		if (m_ViewInfo.lstZoomLevels.indexOf(newZoom) != -1)
		{
			// Use this zoom level
			m_ViewInfo.ZoomLevel = newZoom;
			qDebug() << "Using zoom level " << m_ViewInfo.ZoomLevel;
			break;
		}
		else if (m_bCanScale && (m_ViewInfo.lstZoomLevels.indexOf(newZoom + 1) != -1))
		{
			// Use this zoom level but scale
			m_ViewInfo.ZoomLevel = newZoom;
			qDebug() << "Scaling for zoom level " << m_ViewInfo.ZoomLevel;
			break;
		}
	}
	update();
}

void QMapViewScene::ZoomOut()
{
	quint32 existingZoom = m_ViewInfo.ZoomLevel;

	for (quint32 newZoom=existingZoom-1;newZoom>=m_MinZoom;newZoom--)
	{
		if (m_ViewInfo.lstZoomLevels.indexOf(newZoom) != -1)
		{
			// Use this zoom level
			m_ViewInfo.ZoomLevel = newZoom;
			qDebug() << "Using zoom level " << m_ViewInfo.ZoomLevel;
			break;
		}
		else if (m_bCanScale && (m_ViewInfo.lstZoomLevels.indexOf(newZoom + 1) != -1))
		{
			// Use this zoom level but scale
			m_ViewInfo.ZoomLevel = newZoom;
			qDebug() << "Scaling for zoom level " << m_ViewInfo.ZoomLevel;
			break;
		}
	}
	update();
}

QList<QImage> QMapViewScene::GetImage(quint32 x_index, quint32 y_index)
{
	QList<QImage> lstImages;
	QImage img;
	// First try the real zoom level
	img = GetImage(x_index, y_index, m_ViewInfo.ZoomLevel);
	if (( ! m_bCanScale) || ( ! img.isNull()))
	{
		lstImages.append(img);
		return lstImages;
	}

	// Now try the next zoom level in
	
	quint32 zoomed_x_index = x_index * 2;
	quint32 zoomed_y_index = y_index * 2;

	quint32 x, y;

	for (x=zoomed_x_index;x<=(zoomed_x_index+1);x++)
	{
		for (y=zoomed_y_index;y<=(zoomed_y_index+1);y++)
		{
			img = GetImage(x, y, m_ViewInfo.ZoomLevel + 1);
			if (img.isNull())
			{
				qDebug() << "Null scaled image:" << (m_ViewInfo.ZoomLevel + 1) << "," << x << "," << y;
				lstImages.append(QImage());
			}
			else
			{
				lstImages.append(img.scaledToWidth(m_TileSize/2, Qt::SmoothTransformation));
			}
		}
	}

	return lstImages;
}

/**
 * This function reads a given image from the source data.
 */
QImage QMapViewScene::GetImage(quint32 x_index, quint32 y_index, quint32 zoom)
{
	QImage img;

	if ((pImgReader == NULL) || (! pImgReader->IsOpen()))
		return img;

	img = pImgReader->GetImage(x_index, y_index, zoom);

	return img;

}

void QMapViewScene::SetMapLocation(QString location)
{
	QList<quint32> lstZoomLevels;

	QImage img;
	QFileInfo mapFI(location);
	if ( ! mapFI.exists())
		return;

	if (pImgReader != NULL)
	{
		pImgReader->Close();
		delete pImgReader;
		pImgReader = NULL;
	}

	if ( mapFI.isDir() )
	{
		pImgReader = new MaverickReader();
	}
	else if (location.endsWith(".gemf"))
	{
		pImgReader = new GEMFReader();
	}
	else
	{
		qDebug() << "Unrecognised map store type";
		return;
	}

	pImgReader->Initialise(location);
	if ( ! pImgReader->IsOpen())
		return;

	lstZoomLevels = pImgReader->GetZoomLevels();

	if (lstZoomLevels.count() == 0)
	{
		m_ViewInfo.ZoomLevel = 0xFFFF;
		return;
	}

	if ( ! lstZoomLevels.contains(m_ViewInfo.ZoomLevel))
	{
		SetZoomLevel(lstZoomLevels.at(0));
	}

	m_ViewInfo.lstZoomLevels = lstZoomLevels;

	m_ViewInfo.MapStore = location;
	
	update();
}

QRectF QMapViewScene::getMarkerBoundingBox(LatLon marker)
{
	const quint32 cross_half_size = 10;

	QPointF marker_centre = Mercator::LatLonToWindowPos(marker,
			m_ViewInfo.PageCentre, size(),
			m_ViewInfo.ZoomLevel, m_TileSize);

	QRectF r_marker(marker_centre.x() - cross_half_size,
			marker_centre.y() - cross_half_size,
			cross_half_size * 2,
			cross_half_size * 2);

	return r_marker;

}

void QMapViewScene::DrawImage(QPainter *painter, QImage img, QPointF corner)
{
	if (img.isNull())
		return;
	painter->drawImage(corner, img);
}
void QMapViewScene::paintEvent(QPaintEvent *event)
{
	QFrame::paintEvent(event);
	QList<QImage> lstImages;

	quint32 tilesize = m_TileSize;

	if ((m_ViewInfo.ZoomLevel < m_MinZoom) || (m_ViewInfo.ZoomLevel > m_MaxZoom))
	{
		return;
	}

	Tile t = Mercator::LatLonToOSMTile(m_ViewInfo.PageCentre,
			m_ViewInfo.ZoomLevel, m_TileSize);

	QPainter painter(this);

	QSize winSize = size();

	QPointF centre(winSize.width()/2.0, winSize.height() /2.0);

	QPointF centre_tile_corner(centre.x() - t.posX, centre.y() - t.posY);

	QPointF start_tile_corner(centre_tile_corner);
	quint32 start_x_index = t.indexX;
	quint32 start_y_index = t.indexY;
	while (start_tile_corner.x() > 0.0)
	{
		start_tile_corner.setX(start_tile_corner.x() - tilesize);
		start_x_index--;
	}
	while (start_tile_corner.y() > 0.0)
	{
		start_tile_corner.setY(start_tile_corner.y() - tilesize);
		start_y_index--;
	}

	QPointF tile_corner(start_tile_corner);
	quint32 x_index = start_x_index;
	quint32 y_index = start_y_index;
	while (tile_corner.y() < winSize.height())
	{
		while (tile_corner.x() < winSize.width())
		{
			lstImages = GetImage(x_index, y_index);
			if (lstImages.count() == 1)
			{
				DrawImage(&painter, lstImages.at(0), tile_corner);
			}
			else if (lstImages.count() == 4)
			{
				QPointF subtile_corner = tile_corner;
				DrawImage(&painter, lstImages.at(0), subtile_corner);
				subtile_corner.setY(tile_corner.y() + (m_TileSize/2));
				DrawImage(&painter, lstImages.at(1), subtile_corner);
				subtile_corner.setX(tile_corner.x() + (m_TileSize/2));
				subtile_corner.setY(tile_corner.y());
				DrawImage(&painter, lstImages.at(2), subtile_corner);
				subtile_corner.setY(tile_corner.y() + (m_TileSize/2));
				DrawImage(&painter, lstImages.at(3), subtile_corner);
			}
			else
			{
			}

			tile_corner.setX(tile_corner.x() + tilesize);
			x_index += 1;
		}

		tile_corner.setY(tile_corner.y() + tilesize);
		tile_corner.setX(start_tile_corner.x());
		x_index = start_x_index;
		y_index += 1;
	}


	QPen markerPen;
	markerPen.setWidth(3);
	painter.setPen(markerPen);

	LatLon l;
	foreach (l, m_lstMarkerLocations)
	{
		QRectF m = getMarkerBoundingBox(l);
		if (m.intersects(rect()))
		{
			painter.drawLine(m.topLeft(), m.bottomRight());
			painter.drawLine(m.topRight(), m.bottomLeft());
		}
	}

	if ( ! m_SelectionRect.isNull())
	{
		QPen selectionPen;
		selectionPen.setWidth(2);
		selectionPen.setStyle(Qt::DashLine);
		painter.setPen(selectionPen);
		painter.drawRect(m_SelectionRect);
	}

	if ( lstCoverage.count() > 0)
	{
		QPen coveragePen(Qt::NoPen);
		QColor coverageColour(Qt::magenta);
		coverageColour.setAlpha(100);
		QBrush coverageBrush(coverageColour);
		painter.setPen(coveragePen);
		painter.setBrush(coverageBrush);

		Coverage block;
		foreach (block, lstCoverage)
		{
			QPointF ca = Mercator::LatLonToWindowPos(block.TopLeft,
					m_ViewInfo.PageCentre,
					winSize,
					m_ViewInfo.ZoomLevel,
					m_TileSize);
			QPointF cb = Mercator::LatLonToWindowPos(block.BottomRight,
					m_ViewInfo.PageCentre,
					winSize,
					m_ViewInfo.ZoomLevel,
					m_TileSize);
			painter.drawRect(QRectF(ca, cb));
		}
	}
}
